from itertools import product

from bs4 import BeautifulSoup

# credit https://stackoverflow.com/questions/48393253/how-to-parse-table-with-rowspan-and-colspan/48451104#48451104


def table_to_2d(table_str: str) -> list[list[str | None]]:
    table_tag = BeautifulSoup(table_str, "html.parser")
    rowspans: list[int] = []  # track pending rowspans
    rows = table_tag.find_all("tr")

    # first scan, see how many columns we need
    colcount = 0
    for r, row in enumerate(rows):
        cells = row.find_all(["td", "th"], recursive=False)
        # count columns (including spanned).
        # add active rowspans from preceding rows
        # we *ignore* the colspan value on the last cell, to prevent
        # creating 'phantom' columns with no actual cells, only extended
        # colspans. This is achieved by hardcoding the last cell width as 1.
        # a colspan of 0 means “fill until the end” but can really only apply
        # to the last cell; ignore it elsewhere.
        colcount = max(
            colcount,
            sum(int(c.get("colspan", 1)) or 1 for c in cells[:-1]) + len(cells[-1:]) + len(rowspans),
        )
        # update rowspan bookkeeping; 0 is a span to the bottom.
        rowspans += [int(c.get("rowspan", 1)) or len(rows) - r for c in cells]
        rowspans = [s - 1 for s in rowspans if s > 1]

    # it doesn't matter if there are still rowspan numbers 'active'; no extra
    # rows to show in the table means the larger than 1 rowspan numbers in the
    # last table row are ignored.

    # build an empty matrix for all possible cells
    table: list[list[str | None]] = [[None] * colcount for row in rows]

    # fill matrix from row data
    rowspans_dict: dict[int, int] = {}  # track pending rowspans, column number mapping to count
    for row, row_elem in enumerate(rows):
        span_offset = 0  # how many columns are skipped due to row and colspans
        for col, cell in enumerate(row_elem.find_all(["td", "th"], recursive=False)):
            # adjust for preceding row and colspans
            col += span_offset
            while rowspans_dict.get(col, 0):
                span_offset += 1
                col += 1

            # fill table data
            rowspan = rowspans_dict[col] = int(cell.get("rowspan", 1)) or len(rows) - row
            colspan = int(cell.get("colspan", 1)) or colcount - col
            # next column is offset by the colspan
            span_offset += colspan - 1
            value = cell.get_text()
            for drow, dcol in product(range(rowspan), range(colspan)):
                try:
                    table[row + drow][col + dcol] = value
                    rowspans_dict[col + dcol] = rowspan
                except IndexError:
                    # rowspan or colspan outside the confines of the table
                    pass

        # update rowspan bookkeeping
        rowspans_dict = {c: s - 1 for c, s in rowspans_dict.items() if s > 1}

    return table



from pathlib import Path

import pytest
from syrupy.assertion import SnapshotAssertion
from syrupy.extensions.amber import AmberSnapshotExtension
from syrupy.location import PyTestLocation

from .utils.ffmpeg import get_ffmpeg_version


@pytest.fixture(scope="session")
def ffmpeg_version() -> str:
    return get_ffmpeg_version("minor")


class DifferentDirectoryExtension(AmberSnapshotExtension):
    @classmethod
    def dirname(cls, *, test_location: "PyTestLocation") -> str:
        version = get_ffmpeg_version("minor")
        return str(Path(test_location.filepath).parent / "__snapshots__" / version)


@pytest.fixture
def snapshot_ffmpeg(snapshot: SnapshotAssertion) -> SnapshotAssertion:
    return snapshot.use_extension(DifferentDirectoryExtension)



import re
from typing import List

import requests
from bs4 import BeautifulSoup
from pydantic import BaseModel

from .table import table_to_2d


class ExtInfo(BaseModel):
    ext: str
    description: str
    used_by: str


def extract_tables_from_wikipedia(html: str) -> List[str]:
    # Create a BeautifulSoup object to parse the HTML content
    soup = BeautifulSoup(html, "html.parser")

    # Find all table elements on the page
    tables = soup.find_all("table", class_="wikitable")

    return [str(k) for k in tables]


urls = [
    "https://en.wikipedia.org/wiki/List_of_filename_extensions_(0%E2%80%939)",
    "https://en.wikipedia.org/wiki/List_of_filename_extensions_(A%E2%80%93E)",
    "https://en.wikipedia.org/wiki/List_of_filename_extensions_(F%E2%80%93L)",
    "https://en.wikipedia.org/wiki/List_of_filename_extensions_(M%E2%80%93R)",
    "https://en.wikipedia.org/wiki/List_of_filename_extensions_(S%E2%80%93Z)",
]


def __remove_note(text: str | None) -> str:
    if text:
        return re.sub(r"\[\d*\]", "", text.strip())
    return ""


def _extract_wiki_ext_info(url: str) -> list[ExtInfo]:
    # Extract tables from the Wikipedia page

    html = requests.get(url).text
    tables = extract_tables_from_wikipedia(html)
    output = []
    for table in tables:
        normalize_table = table_to_2d(table)

        for row in normalize_table[1:]:
            output.append(
                ExtInfo(
                    ext=__remove_note(row[0]),
                    description=__remove_note(row[1]),
                    used_by=__remove_note(row[2]),
                )
            )

    return output


def extract_wiki_filename_exts() -> list[ExtInfo]:
    output = []
    for url in urls:
        output.extend(_extract_wiki_ext_info(url))

    return output



import os
from typing import Literal
from urllib.parse import urlparse

from pydantic import BaseModel

from .utils.ffmpeg import FFProbeInfo, ffprobe, get_ffmpeg, load_cache
from .utils.shell import call


class MediaInfo(BaseModel):
    type: Literal["image", "video", "audio"]

    width: int | None = None
    height: int | None = None
    duration: float | None = None

    format: str | None = None
    size: int | None = None
    suggest_ext: str | None = None


def generate_thumbnail(video_path: str, thumbnail_path: str, time_offset: float = 0) -> str:
    ffmpeg_cmd = get_ffmpeg() + [
        "-y",  # Overwrite output file if it exists
        "-i",
        video_path,  # Input video path
        "-ss",
        str(time_offset),  # Time offset (seek to the specified position)
        "-vframes",
        "1",  # Number of frames to output
        "-vf",
        "scale=320:-1",  # Thumbnail size (width: 320, height: proportional)
        "-q:v",
        "2",  # Quality (2 - high, 5 - low)
        thumbnail_path,  # Output thumbnail path
    ]

    call(ffmpeg_cmd)

    return thumbnail_path


def _extract_file_extension(uri: str) -> str:
    parsed_uri = urlparse(uri)
    path = parsed_uri.path
    filename = os.path.basename(path)
    _, extension = os.path.splitext(filename)
    return extension[1:].lower()


_KNOWN_CODEC_EXTS = {
    "png_pipe": ["png"],
    "svg_pipe": ["svg"],
    "tiff_pipe": ["tiff", "tif"],
    "bmp_pipe": ["bmp"],
    "gif_pipe": ["gif"],
    "jpeg_pipe": ["jpeg", "jpg"],
    "webp_pipe": ["webp"],
    "mjpeg": ["jpg", "jpeg", "mjpeg"],
}


def _guess_media_info(
    uri: str,
    info: FFProbeInfo,
) -> MediaInfo:
    infos = load_cache()
    current_ext = _extract_file_extension(uri)

    format_name = info.format.format_name
    duration = info.format.duration

    # NOTE: handle ffmpeg's image compatibility
    if format_name == "image2":
        format_name = info.streams[0].codec_name

    # NOTE: detect file extension
    if format_name in _KNOWN_CODEC_EXTS:
        common_exts = _KNOWN_CODEC_EXTS[format_name]
    elif format_name in infos:
        common_exts = infos[format_name].common_exts
    else:
        common_exts = []

    if current_ext in common_exts:
        suggest_ext = current_ext
    elif common_exts:
        suggest_ext = common_exts[0]
    else:
        suggest_ext = None

    # NOTE: we classify gif as image
    if not duration or format_name in ("gif", "mjpeg"):
        return MediaInfo(
            type="image",
            width=info.streams[0].width or 0,
            height=info.streams[0].height or 0,
            duration=duration or 0,
            format=format_name,
            size=info.format.size,
            suggest_ext=suggest_ext,
        )

    for stream in info.streams:
        # NOTE: if there is at least one video stream, the media is video
        if stream.codec_type == "video":
            width = stream.width
            height = stream.height

            return MediaInfo(
                type="video",
                width=width or 0,
                height=height or 0,
                duration=duration or 0,
                format=format_name,
                size=info.format.size,
                suggest_ext=suggest_ext,
            )

    # NOTE: if there is no video stream, the media is audio
    return MediaInfo(
        type="audio",
        width=0,
        height=0,
        duration=duration or 0,
        format=format_name,
        size=info.format.size,
        suggest_ext=suggest_ext,
    )


def detect(uri: str) -> MediaInfo:
    probe_info = ffprobe(uri)
    return _guess_media_info(uri, probe_info)
